# Copyright (C) 1998-2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

import os
import md5
import sys
import time
import email
import errno
import shutil
import cPickle
import marshal
import optparse

from flufl.lock import TimeOutError

from mailman import MailList
from mailman import Message
from mailman import Pending
from mailman import Utils
from mailman import version
from mailman.MemberAdaptor import BYBOUNCE, ENABLED
from mailman.OldStyleMemberships import OldStyleMemberships
from mailman.Queue.Switchboard import Switchboard
from mailman.configuration import config
from mailman.core.i18n import _
from mailman.initialize import initialize
from mailman.utilities.filesystem import makedirs


FRESH = 0
NOTFRESH = -1



def parseargs():
    parser = optparse.OptionParser(version=version.MAILMAN_VERSION,
                                   usage=_("""\
Perform all necessary upgrades.

%prog [options]"""))
    parser.add_option('-f', '--force',
                      default=False, action='store_true', help=_("""\
Force running the upgrade procedures.  Normally, if the version number of the
installed Mailman matches the current version number (or a 'downgrade' is
detected), nothing will be done."""))
    parser.add_option('-C', '--config',
                      help=_('Alternative configuration file to use'))
    opts, args = parser.parse_args()
    if args:
        parser.print_help()
        print >> sys.stderr, _('Unexpected arguments')
        sys.exit(1)
    return parser, opts, args



def calcversions():
    # Returns a tuple of (lastversion, thisversion).  If the last version
    # could not be determined, lastversion will be FRESH or NOTFRESH,
    # depending on whether this installation appears to be fresh or not.  The
    # determining factor is whether there are files in the $var_prefix/logs
    # subdir or not.  The version numbers are HEX_VERSIONs.
    #
    # See if we stored the last updated version
    lastversion = None
    thisversion = version.HEX_VERSION
    try:
        fp = open(os.path.join(config.DATA_DIR, 'last_mailman_version'))
        data = fp.read()
        fp.close()
        lastversion = int(data, 16)
    except (IOError, ValueError):
        pass
    #
    # try to figure out if this is a fresh install
    if lastversion is None:
        lastversion = FRESH
        try:
            if os.listdir(config.LOG_DIR):
                lastversion = NOTFRESH
        except OSError:
            pass
    return (lastversion, thisversion)



def makeabs(relpath):
    return os.path.join(config.PREFIX, relpath)


def make_varabs(relpath):
    return os.path.join(config.VAR_PREFIX, relpath)



def move_language_templates(mlist):
    listname = mlist.internal_name()
    print _('Fixing language templates: $listname')
    # Mailman 2.1 has a new cascading search for its templates, defined and
    # described in Utils.py:maketext().  Putting templates in the top level
    # templates/ subdir or the lists/<listname> subdir is deprecated and no
    # longer searched..
    #
    # What this means is that most templates can live in the global templates/
    # subdirectory, and only needs to be copied into the list-, vhost-, or
    # site-specific language directories when needed.
    #
    # Also, by default all standard (i.e. English) templates must now live in
    # the templates/en directory.  This update cleans up all the templates,
    # deleting more-specific duplicates (as calculated by md5 checksums) in
    # favor of more-global locations.
    #
    # First, get rid of any lists/<list> template or lists/<list>/en template
    # that is identical to the global templates/* default.
    for gtemplate in os.listdir(os.path.join(config.TEMPLATE_DIR, 'en')):
        # BAW: get rid of old templates, e.g. admlogin.txt and
        # handle_opts.html
        try:
            fp = open(os.path.join(config.TEMPLATE_DIR, gtemplate))
        except IOError, e:
            if e.errno <> errno.ENOENT:
                raise
            # No global template
            continue
        gcksum = md5.new(fp.read()).digest()
        fp.close()
        # Match against the lists/<list>/* template
        try:
            fp = open(os.path.join(mlist.fullpath(), gtemplate))
        except IOError, e:
            if e.errno <> errno.ENOENT:
                raise
        else:
            tcksum = md5.new(fp.read()).digest()
            fp.close()
            if gcksum == tcksum:
                os.unlink(os.path.join(mlist.fullpath(), gtemplate))
        # Match against the lists/<list>/*.prev template
        try:
            fp = open(os.path.join(mlist.fullpath(), gtemplate + '.prev'))
        except IOError, e:
            if e.errno <> errno.ENOENT:
                raise
        else:
            tcksum = md5.new(fp.read()).digest()
            fp.close()
            if gcksum == tcksum:
                os.unlink(os.path.join(mlist.fullpath(), gtemplate + '.prev'))
        # Match against the lists/<list>/en/* templates
        try:
            fp = open(os.path.join(mlist.fullpath(), 'en', gtemplate))
        except IOError, e:
            if e.errno <> errno.ENOENT:
                raise
        else:
            tcksum = md5.new(fp.read()).digest()
            fp.close()
            if gcksum == tcksum:
                os.unlink(os.path.join(mlist.fullpath(), 'en', gtemplate))
        # Match against the templates/* template
        try:
            fp = open(os.path.join(config.TEMPLATE_DIR, gtemplate))
        except IOError, e:
            if e.errno <> errno.ENOENT:
                raise
        else:
            tcksum = md5.new(fp.read()).digest()
            fp.close()
            if gcksum == tcksum:
                os.unlink(os.path.join(config.TEMPLATE_DIR, gtemplate))
        # Match against the templates/*.prev template
        try:
            fp = open(os.path.join(config.TEMPLATE_DIR, gtemplate + '.prev'))
        except IOError, e:
            if e.errno <> errno.ENOENT:
                raise
        else:
            tcksum = md5.new(fp.read()).digest()
            fp.close()
            if gcksum == tcksum:
                os.unlink(os.path.join(config.TEMPLATE_DIR,
                                       gtemplate + '.prev'))



def situate_list(listname):
    # This turns the directory called 'listname' into a directory called
    # 'listname@domain'.  Start by finding out what the domain should be.
    # A list's domain is its email host.
    mlist = MailList.MailList(listname, lock=False, check_version=False)
    fullname = mlist.fqdn_listname
    oldpath = os.path.join(config.VAR_PREFIX, 'lists', listname)
    newpath = os.path.join(config.VAR_PREFIX, 'lists', fullname)
    if os.path.exists(newpath):
        print >> sys.stderr, _('WARNING: could not situate list: $listname')
    else:
        os.rename(oldpath, newpath)
        print _('situated list $listname to $fullname')
    return fullname



def dolist(listname):
    mlist = MailList.MailList(listname, lock=False)
    try:
        mlist.Lock(0.5)
    except TimeOutError:
        print >> sys.stderr, _(
            'WARNING: could not acquire lock for list: $listname')
        return 1
    # Sanity check the invariant that every BYBOUNCE disabled member must have
    # bounce information.  Some earlier betas broke this.  BAW: we're
    # submerging below the MemberAdaptor interface, so skip this if we're not
    # using OldStyleMemberships.
    if isinstance(mlist._memberadaptor, OldStyleMemberships):
        noinfo = {}
        for addr, (reason, when) in mlist.delivery_status.items():
            if reason == BYBOUNCE and not mlist.bounce_info.has_key(addr):
                noinfo[addr] = reason, when
        # What to do about these folks with a BYBOUNCE delivery status and no
        # bounce info?  This number should be very small, and I think it's
        # fine to simple re-enable them and let the bounce machinery
        # re-disable them if necessary.
        n = len(noinfo)
        if n > 0:
            print _(
                'Resetting $n BYBOUNCEs disabled addrs with no bounce info')
            for addr in noinfo.keys():
                mlist.setDeliveryStatus(addr, ENABLED)

    mbox_dir = make_varabs('archives/private/%s.mbox' % (listname))
    mbox_file = make_varabs('archives/private/%s.mbox/%s' % (listname,
                                                             listname))
    o_pub_mbox_file = make_varabs('archives/public/%s' % (listname))
    o_pri_mbox_file = make_varabs('archives/private/%s' % (listname))
    html_dir = o_pri_mbox_file
    o_html_dir = makeabs('public_html/archives/%s' % (listname))
    # Make the mbox directory if it's not there.
    if not os.path.exists(mbox_dir):
        makedirs(mbox_dir)
    else:
        # This shouldn't happen, but hey, just in case
        if not os.path.isdir(mbox_dir):
            print _("""\
For some reason, $mbox_dir exists as a file.  This won't work with b6, so I'm
renaming it to ${mbox_dir}.tmp and proceeding.""")
            os.rename(mbox_dir, "%s.tmp" % (mbox_dir))
            makedirs(mbox_dir)
    # Move any existing mboxes around, but watch out for both a public and a
    # private one existing
    if os.path.isfile(o_pri_mbox_file) and os.path.isfile(o_pub_mbox_file):
        if mlist.archive_private:
            print _("""\

$listname has both public and private mbox archives.  Since this list
currently uses private archiving, I'm installing the private mbox archive --
$o_pri_mbox_file -- as the active archive, and renaming
        $o_pub_mbox_file
to
        ${o_pub_mbox_file}.preb6

You can integrate that into the archives if you want by using the 'arch'
script.
""") % (mlist._internal_name, o_pri_mbox_file, o_pub_mbox_file,
        o_pub_mbox_file)
            os.rename(o_pub_mbox_file, "%s.preb6" % (o_pub_mbox_file))
        else:
            print _("""\
$mlist._internal_name has both public and private mbox archives.  Since this
list currently uses public archiving, I'm installing the public mbox file
archive file ($o_pub_mbox_file) as the active one, and renaming
$o_pri_mbox_file to ${o_pri_mbox_file}.preb6

You can integrate that into the archives if you want by using the 'arch'
script.
""")
            os.rename(o_pri_mbox_file, "%s.preb6" % (o_pri_mbox_file))
    # Move private archive mbox there if it's around
    # and take into account all sorts of absurdities
    print _('- updating old private mbox file')
    if os.path.exists(o_pri_mbox_file):
        if os.path.isfile(o_pri_mbox_file):
            os.rename(o_pri_mbox_file, mbox_file)
        elif not os.path.isdir(o_pri_mbox_file):
            newname = "%s.mm_install-dunno_what_this_was_but_its_in_the_way" \
                      % o_pri_mbox_file
            os.rename(o_pri_mbox_file, newname)
            print _("""\
    unknown file in the way, moving
        $o_pri_mbox_file
    to
        $newname""")
        else:
            # directory
            print _("""\
    looks like you have a really recent development installation...
    you're either one brave soul, or you already ran me""")
    # Move public archive mbox there if it's around
    # and take into account all sorts of absurdities.
    print _('- updating old public mbox file')
    if os.path.exists(o_pub_mbox_file):
        if os.path.isfile(o_pub_mbox_file):
            os.rename(o_pub_mbox_file, mbox_file)
        elif not os.path.isdir(o_pub_mbox_file):
            newname = "%s.mm_install-dunno_what_this_was_but_its_in_the_way" \
                      % o_pub_mbox_file
            os.rename(o_pub_mbox_file, newname)
            print _("""\
    unknown file in the way, moving
        $o_pub_mbox_file
    to
        $newname""")
        else: # directory
            print _("""\
    looks like you have a really recent development installation...
    you're either one brave soul, or you already ran me""")
    # Move the html archives there
    if os.path.isdir(o_html_dir):
        os.rename(o_html_dir, html_dir)
        # chmod the html archives
        os.chmod(html_dir, 02775)
    # BAW: Is this still necessary?!
    mlist.Save()
    # Check to see if pre-b4 list-specific templates are around
    # and move them to the new place if there's not already
    # a new one there
    tmpl_dir = os.path.join(config.PREFIX, "templates")
    list_dir = os.path.join(config.PREFIX, "lists")
    b4_tmpl_dir = os.path.join(tmpl_dir, mlist._internal_name)
    new_tmpl_dir = os.path.join(list_dir, mlist._internal_name)
    if os.path.exists(b4_tmpl_dir):
        print _("""\
- This list looks like it might have <= b4 list templates around""")
        for f in os.listdir(b4_tmpl_dir):
            o_tmpl = os.path.join(b4_tmpl_dir, f)
            n_tmpl = os.path.join(new_tmpl_dir, f)
            if os.path.exists(o_tmpl):
                if not os.path.exists(n_tmpl):
                    os.rename(o_tmpl, n_tmpl)
                    print _('- moved $o_tmpl to $n_tmpl')
                else:
                    print _("""\
- both $o_tmpl and $n_tmpl exist, leaving untouched""")
            else:
                print _("""\
- $o_tmpl doesn't exist, leaving untouched""")
    # Move all the templates to the en language subdirectory as required for
    # Mailman 2.1
    move_language_templates(mlist)
    # Avoid eating filehandles with the list lockfiles
    mlist.Unlock()
    return 0



def archive_path_fixer(unused_arg, dir, files):
    # Passed to os.path.walk to fix the perms on old html archives.
    for f in files:
        abs = os.path.join(dir, f)
        if os.path.isdir(abs):
            if f == "database":
                os.chmod(abs, 02770)
            else:
                os.chmod(abs, 02775)
        elif os.path.isfile(abs):
            os.chmod(abs, 0664)


def remove_old_sources(module):
    # Also removes old directories.
    src = '%s/%s' % (config.PREFIX, module)
    pyc = src + "c"
    if os.path.isdir(src):
        print _('removing directory $src and everything underneath')
        shutil.rmtree(src)
    elif os.path.exists(src):
        print _('removing $src')
        try:
            os.unlink(src)
        except os.error, rest:
            print _("Warning: couldn't remove $src -- $rest")
    if module.endswith('.py') and os.path.exists(pyc):
        try:
            os.unlink(pyc)
        except OSError, rest:
            print _("couldn't remove old file $pyc -- $rest")



def update_qfiles():
    print _('updating old qfiles')
    prefix = `time.time()` + '+'
    # Be sure the qfiles/in directory exists (we don't really need the
    # switchboard object, but it's convenient for creating the directory).
    sb = Switchboard(config.INQUEUE_DIR)
    for filename in os.listdir(config.QUEUE_DIR):
        # Updating means just moving the .db and .msg files to qfiles/in where
        # it should be dequeued, converted, and processed normally.
        if os.path.splitext(filename) == '.msg':
            oldmsgfile = os.path.join(config.QUEUE_DIR, filename)
            newmsgfile = os.path.join(config.INQUEUE_DIR, prefix + filename)
            os.rename(oldmsgfile, newmsgfile)
        elif os.path.splitext(filename) == '.db':
            olddbfile = os.path.join(config.QUEUE_DIR, filename)
            newdbfile = os.path.join(config.INQUEUE_DIR, prefix + filename)
            os.rename(olddbfile, newdbfile)
    # Now update for the Mailman 2.1.5 qfile format.  For every filebase in
    # the qfiles/* directories that has both a .pck and a .db file, pull the
    # data out and re-queue them.
    for dirname in os.listdir(config.QUEUE_DIR):
        dirpath = os.path.join(config.QUEUE_DIR, dirname)
        if dirpath == config.BADQUEUE_DIR:
            # The files in qfiles/bad can't possibly be pickles
            continue
        sb = Switchboard(dirpath)
        try:
            for filename in os.listdir(dirpath):
                filepath = os.path.join(dirpath, filename)
                filebase, ext = os.path.splitext(filepath)
                # Handle the .db metadata files as part of the handling of the
                # .pck or .msg message files.
                if ext not in ('.pck', '.msg'):
                    continue
                msg, data = dequeue(filebase)
                if msg is not None and data is not None:
                    sb.enqueue(msg, data)
        except EnvironmentError, e:
            if e.errno <> errno.ENOTDIR:
                raise
            print _('Warning!  Not a directory: $dirpath')



# Implementations taken from the pre-2.1.5 Switchboard
def ext_read(filename):
    fp = open(filename)
    d = marshal.load(fp)
    # Update from version 2 files
    if d.get('version', 0) == 2:
        del d['filebase']
    # Do the reverse conversion (repr -> float)
    for attr in ['received_time']:
        try:
            sval = d[attr]
        except KeyError:
            pass
        else:
            # Do a safe eval by setting up a restricted execution
            # environment.  This may not be strictly necessary since we
            # know they are floats, but it can't hurt.
            d[attr] = eval(sval, {'__builtins__': {}})
    fp.close()
    return d


def dequeue(filebase):
    # Calculate the .db and .msg filenames from the given filebase.
    msgfile = os.path.join(filebase + '.msg')
    pckfile = os.path.join(filebase + '.pck')
    dbfile = os.path.join(filebase + '.db')
    # Now we are going to read the message and metadata for the given
    # filebase.  We want to read things in this order: first, the metadata
    # file to find out whether the message is stored as a pickle or as
    # plain text.  Second, the actual message file.  However, we want to
    # first unlink the message file and then the .db file, because the
    # qrunner only cues off of the .db file
    msg = None
    try:
        data = ext_read(dbfile)
        os.unlink(dbfile)
    except EnvironmentError, e:
        if e.errno <> errno.ENOENT:
            raise
        data = {}
    # Between 2.1b4 and 2.1b5, the `rejection-notice' key in the metadata
    # was renamed to `rejection_notice', since dashes in the keys are not
    # supported in METAFMT_ASCII.
    if data.has_key('rejection-notice'):
        data['rejection_notice'] = data['rejection-notice']
        del data['rejection-notice']
    msgfp = None
    try:
        try:
            msgfp = open(pckfile)
            msg = cPickle.load(msgfp)
            os.unlink(pckfile)
        except EnvironmentError, e:
            if e.errno <> errno.ENOENT: raise
            msgfp = None
            try:
                msgfp = open(msgfile)
                msg = email.message_from_file(msgfp, Message.Message)
                os.unlink(msgfile)
            except EnvironmentError, e:
                if e.errno <> errno.ENOENT: raise
            except (email.Errors.MessageParseError, ValueError), e:
                # This message was unparsable, most likely because its
                # MIME encapsulation was broken.  For now, there's not
                # much we can do about it.
                print _('message is unparsable: $filebase')
                msgfp.close()
                msgfp = None
                if config.QRUNNER_SAVE_BAD_MESSAGES:
                    # Cheapo way to ensure the directory exists w/ the
                    # proper permissions.
                    sb = Switchboard(config.BADQUEUE_DIR)
                    os.rename(msgfile, os.path.join(
                        config.BADQUEUE_DIR, filebase + '.txt'))
                else:
                    os.unlink(msgfile)
                msg = data = None
        except EOFError:
            # For some reason the pckfile was empty.  Just delete it.
            print _('Warning!  Deleting empty .pck file: $pckfile')
            os.unlink(pckfile)
    finally:
        if msgfp:
            msgfp.close()
    return msg, data



def main():
    parser, opts, args = parseargs()
    initialize(opts.config)

    # calculate the versions
    lastversion, thisversion = calcversions()
    hexlversion = hex(lastversion)
    hextversion = hex(thisversion)
    if lastversion == thisversion and not opts.force:
        # nothing to do
        print _('No updates are necessary.')
        sys.exit(0)
    if lastversion > thisversion and not opts.force:
        print _("""\
Downgrade detected, from version $hexlversion to version $hextversion
This is probably not safe.
Exiting.""")
        sys.exit(1)
    print _('Upgrading from version $hexlversion to $hextversion')
    errors = 0
    # get rid of old stuff
    print _('getting rid of old source files')
    for mod in ('mailman/Archiver.py', 'mailman/HyperArch.py',
                'mailman/HyperDatabase.py', 'mailman/pipermail.py',
                'mailman/smtplib.py', 'mailman/Cookie.py',
                'bin/update_to_10b6', 'scripts/mailcmd',
                'scripts/mailowner', 'mail/wrapper', 'mailman/pythonlib',
                'cgi-bin/archives', 'mailman/MailCommandHandler'):
        remove_old_sources(mod)
    if not config.list_manager.names:
        print _('no lists == nothing to do, exiting')
        return
    # For people with web archiving, make sure the directories
    # in the archiving are set with proper perms for b6.
    if os.path.isdir("%s/public_html/archives" % config.PREFIX):
        print _("""\
fixing all the perms on your old html archives to work with b6
If your archives are big, this could take a minute or two...""")
        os.path.walk("%s/public_html/archives" % config.PREFIX,
                     archive_path_fixer, "")
        print _('done')
    for listname in config.list_manager.names:
        # With 2.2.0a0, all list names grew an @domain suffix.  If you find a
        # list without that, move it now.
        if not '@' in listname:
            listname = situate_list(listname)
        print _('Updating mailing list: $listname')
        errors += dolist(listname)
        print
    print _('Updating Usenet watermarks')
    wmfile = os.path.join(config.DATA_DIR, 'gate_watermarks')
    try:
        fp = open(wmfile)
    except IOError:
        print _('- nothing to update here')
    else:
        d = marshal.load(fp)
        fp.close()
        for listname in d.keys():
            if listname not in listnames:
                # this list no longer exists
                continue
            mlist = MailList.MailList(listname, lock=0)
            try:
                mlist.Lock(0.5)
            except TimeOutError:
                print >> sys.stderr, _(
                    'WARNING: could not acquire lock for list: $listname')
                errors = errors + 1
            else:
                # Pre 1.0b7 stored 0 in the gate_watermarks file to indicate
                # that no gating had been done yet.  Without coercing this to
                # None, the list could now suddenly get flooded.
                mlist.usenet_watermark = d[listname] or None
                mlist.Save()
                mlist.Unlock()
        os.unlink(wmfile)
        print _('- usenet watermarks updated and gate_watermarks removed')
    # In Mailman 2.1, the qfiles directory has a different structure and a
    # different content.  Also, in Mailman 2.1.5 we collapsed the message
    # files from separate .msg (pickled Message objects) and .db (marshalled
    # dictionaries) to a shared .pck file containing two pickles.
    update_qfiles()
    # This warning was necessary for the upgrade from 1.0b9 to 1.0b10.
    # There's no good way of figuring this out for releases prior to 2.0beta2
    # :(
    if lastversion == NOTFRESH:
        print _("""

NOTE NOTE NOTE NOTE NOTE

    You are upgrading an existing Mailman installation, but I can't tell what
    version you were previously running.

    If you are upgrading from Mailman 1.0b9 or earlier you will need to
    manually update your mailing lists.  For each mailing list you need to
    copy the file templates/options.html lists/<listname>/options.html.

    However, if you have edited this file via the Web interface, you will have
    to merge your changes into this file, otherwise you will lose your
    changes.

NOTE NOTE NOTE NOTE NOTE

""")
    if not errors:
        # Record the version we just upgraded to
        fp = open(os.path.join(config.DATA_DIR, 'last_mailman_version'), 'w')
        fp.write(hex(config.HEX_VERSION) + '\n')
        fp.close()
    else:
        lockdir = config.LOCK_DIR
        print _('''\

ERROR:

The locks for some lists could not be acquired.  This means that either
Mailman was still active when you upgraded, or there were stale locks in the
$lockdir directory.

You must put Mailman into a quiescent state and remove all stale locks, then
re-run "make update" manually.  See the INSTALL and UPGRADE files for details.
''')



# Copyright (C) 2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""REST API for Message moderation."""

from __future__ import absolute_import, print_function, unicode_literals

__metaclass__ = type
__all__ = [
    'HeldMessage',
    'HeldMessages',
    ]


from restish import http, resource
from zope.component import getUtility

from mailman.app.moderator import handle_message
from mailman.interfaces.action import Action
from mailman.interfaces.messages import IMessageStore
from mailman.interfaces.requests import IListRequests, RequestType
from mailman.rest.helpers import CollectionMixin, etag, no_content
from mailman.rest.validator import Validator, enum_validator



class HeldMessage(resource.Resource, CollectionMixin):
    """Resource for moderating a held message."""

    def __init__(self, mlist, request_id):
        self._mlist = mlist
        self._request_id = request_id

    @resource.GET()
    def details(self, request):
        requests = IListRequests(self._mlist)
        try:
            request_id = int(self._request_id)
        except ValueError:
            return http.bad_request()
        results = requests.get_request(request_id, RequestType.held_message)
        if results is None:
            return http.not_found()
        key, data = results
        msg = getUtility(IMessageStore).get_message_by_id(key)
        resource = dict(
            key=key,
            data=data,
            msg=msg.as_string(),
            id=request_id,
            )
        return http.ok([], etag(resource))

    @resource.POST()
    def moderate(self, request):
        try:
            validator = Validator(action=enum_validator(Action))
            arguments = validator(request)
        except ValueError as error:
            return http.bad_request([], str(error))
        requests = IListRequests(self._mlist)
        try:
            request_id = int(self._request_id)
        except ValueError:
            return http.bad_request()
        results = requests.get_request(request_id, RequestType.held_message)
        if results is None:
            return http.not_found()
        handle_message(self._mlist, request_id, **arguments)
        return no_content()



class HeldMessages(resource.Resource, CollectionMixin):
    """Resource for messages held for moderation."""

    def __init__(self, mlist):
        self._mlist = mlist

    def _resource_as_dict(self, req):
        """See `CollectionMixin`."""
        key, data = self._requests.get_request(req.id)
        return dict(
            key=key,
            data=data,
            id=req.id,
            )

    def _get_collection(self, request):
        requests = IListRequests(self._mlist)
        self._requests = requests
        return list(requests.of_type(RequestType.held_message))

    @resource.GET()
    def requests(self, request):
        """/lists/listname/held"""
        resource = self._make_collection(request)
        return http.ok([], etag(resource))

    @resource.child('{id}')
    def message(self, request, segments, **kw):
        return HeldMessage(self._mlist, kw['id'])



# Copyright (C) 2007-2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""A mailing list manager."""

from __future__ import absolute_import, unicode_literals

__metaclass__ = type
__all__ = [
    'ListManager',
    ]


import datetime

from zope.event import notify
from zope.interface import implements

from mailman.config import config
from mailman.interfaces.address import InvalidEmailAddressError
from mailman.interfaces.listmanager import (
    IListManager, ListAlreadyExistsError, ListCreatedEvent, ListCreatingEvent,
    ListDeletedEvent, ListDeletingEvent)
from mailman.model.mailinglist import MailingList



class ListManager:
    """An implementation of the `IListManager` interface."""

    implements(IListManager)

    def create(self, fqdn_listname):
        """See `IListManager`."""
        listname, at, hostname = fqdn_listname.partition('@')
        if len(hostname) == 0:
            raise InvalidEmailAddressError(fqdn_listname)
        notify(ListCreatingEvent(fqdn_listname))
        mlist = config.db.store.find(
            MailingList,
            MailingList.list_name == listname,
            MailingList.mail_host == hostname).one()
        if mlist:
            raise ListAlreadyExistsError(fqdn_listname)
        mlist = MailingList(fqdn_listname)
        mlist.created_at = datetime.datetime.now()
        config.db.store.add(mlist)
        notify(ListCreatedEvent(mlist))
        return mlist

    def get(self, fqdn_listname):
        """See `IListManager`."""
        listname, at, hostname = fqdn_listname.partition('@')
        return config.db.store.find(MailingList,
                                    list_name=listname,
                                    mail_host=hostname).one()

    def delete(self, mlist):
        """See `IListManager`."""
        fqdn_listname = mlist.fqdn_listname
        notify(ListDeletingEvent(mlist))
        config.db.store.remove(mlist)
        notify(ListDeletedEvent(fqdn_listname))

    @property
    def mailing_lists(self):
        """See `IListManager`."""
        for mlist in config.db.store.find(MailingList):
            yield mlist

    def __iter__(self):
        """See `IListManager`."""
        for mlist in config.db.store.find(MailingList):
            yield mlist

    @property
    def names(self):
        """See `IListManager`."""
        result_set = config.db.store.find(MailingList)
        for mail_host, list_name in result_set.values(MailingList.mail_host, 
                                                      MailingList.list_name):
            yield '{0}@{1}'.format(list_name, mail_host)

    @property
    def name_components(self):
        """See `IListManager`."""
        result_set = config.db.store.find(MailingList)
        for mail_host, list_name in result_set.values(MailingList.mail_host, 
                                                      MailingList.list_name):
            yield list_name, mail_host



# Copyright (C) 1998-2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""Decorate a message by sticking the header and footer around it."""

from __future__ import absolute_import, print_function, unicode_literals

__metaclass__ = type
__all__ = [
    'Decorate',
    ]


import re
import logging

from email.mime.text import MIMEText
from urllib2 import URLError
from zope.component import getUtility
from zope.interface import implements

from mailman.core.i18n import _
from mailman.email.message import Message
from mailman.interfaces.handler import IHandler
from mailman.interfaces.templates import ITemplateLoader
from mailman.utilities.string import expand


log = logging.getLogger('mailman.error')



def process(mlist, msg, msgdata):
    """Decorate the message with headers and footers."""
    # Digests and Mailman-craft messages should not get additional headers.
    if msgdata.get('isdigest') or msgdata.get('nodecorate'):
        return
    d = {}
    member = msgdata.get('member')
    if member is not None:
        # Calculate the extra personalization dictionary.
        recipient = msgdata.get('recipient', member.address.original_email)
        d['user_address'] = recipient
        d['user_delivered_to'] = member.address.original_email
        d['user_language'] = member.preferred_language.description
        d['user_name'] = (member.user.display_name 
                          if member.user.display_name
                          else member.address.original_email)
        d['user_optionsurl'] = member.options_url
    # These strings are descriptive for the log file and shouldn't be i18n'd
    d.update(msgdata.get('decoration-data', {}))
    try:
        header = decorate(mlist, mlist.header_uri, d)
    except URLError:
        log.exception('Header decorator URI not found ({0}): {1}'.format(
            mlist.fqdn_listname, mlist.header_uri))
    try:
        footer = decorate(mlist, mlist.footer_uri, d)
    except URLError:
        log.exception('Footer decorator URI not found ({0}): {1}'.format(
            mlist.fqdn_listname, mlist.footer_uri))
    # Escape hatch if both the footer and header are empty
    if not header and not footer:
        return
    # Be MIME smart here.  We only attach the header and footer by
    # concatenation when the message is a non-multipart of type text/plain.
    # Otherwise, if it is not a multipart, we make it a multipart, and then we
    # add the header and footer as text/plain parts.
    #
    # BJG: In addition, only add the footer if the message's character set
    # matches the charset of the list's preferred language.  This is a
    # suboptimal solution, and should be solved by allowing a list to have
    # multiple headers/footers, for each language the list supports.
    #
    # Also, if the list's preferred charset is us-ascii, we can always
    # safely add the header/footer to a plain text message since all
    # charsets Mailman supports are strict supersets of us-ascii --
    # no, UTF-16 emails are not supported yet.
    #
    # TK: Message with 'charset=' cause trouble. So, instead of
    #     mgs.get_content_charset('us-ascii') ...
    mcset = msg.get_content_charset() or 'us-ascii'
    lcset = mlist.preferred_language.charset
    msgtype = msg.get_content_type()
    # BAW: If the charsets don't match, should we add the header and footer by
    # MIME multipart chroming the message?
    wrap = True
    if not msg.is_multipart() and msgtype == 'text/plain':
        # Save the RFC-3676 format parameters.
        format_param = msg.get_param('format')
        delsp = msg.get_param('delsp')
        # Save 'Content-Transfer-Encoding' header in case decoration fails.
        cte = msg.get('content-transfer-encoding')
        # header/footer is now in unicode (2.2)
        try:
            oldpayload = unicode(msg.get_payload(decode=True), mcset)
            del msg['content-transfer-encoding']
            frontsep = endsep = ''
            if header and not header.endswith('\n'):
                frontsep = '\n'
            if footer and not oldpayload.endswith('\n'):
                endsep = '\n'
            payload = header + frontsep + oldpayload + endsep + footer
            # When setting the payload for the message, try various charset
            # encodings until one does not produce a UnicodeError.  We'll try
            # charsets in this order: the list's charset, the message's
            # charset, then utf-8.  It's okay if some of these are duplicates.
            for cset in (lcset, mcset, 'utf-8'):
                try:
                    msg.set_payload(payload.encode(cset), cset)
                except UnicodeError:
                    pass
                else:
                    if format_param:
                        msg.set_param('format', format_param)
                    if delsp:
                        msg.set_param('delsp', delsp)
                    wrap = False
                    break
        except (LookupError, UnicodeError):
            if cte:
                # Restore the original c-t-e.
                del msg['content-transfer-encoding']
                msg['Content-Transfer-Encoding'] = cte
    elif msg.get_content_type() == 'multipart/mixed':
        # The next easiest thing to do is just prepend the header and append
        # the footer as additional subparts
        payload = msg.get_payload()
        if not isinstance(payload, list):
            payload = [payload]
        if footer:
            mimeftr = MIMEText(footer.encode(lcset), 'plain', lcset)
            mimeftr['Content-Disposition'] = 'inline'
            payload.append(mimeftr)
        if header:
            mimehdr = MIMEText(header.encode(lcset), 'plain', lcset)
            mimehdr['Content-Disposition'] = 'inline'
            payload.insert(0, mimehdr)
        msg.set_payload(payload)
        wrap = False
    # If we couldn't add the header or footer in a less intrusive way, we can
    # at least do it by MIME encapsulation.  We want to keep as much of the
    # outer chrome as possible.
    if not wrap:
        return
    # Because of the way Message objects are passed around to process(), we
    # need to play tricks with the outer message -- i.e. the outer one must
    # remain the same instance.  So we're going to create a clone of the outer
    # message, with all the header chrome intact, then copy the payload to it.
    # This will give us a clone of the original message, and it will form the
    # basis of the interior, wrapped Message.
    inner = Message()
    # Which headers to copy?  Let's just do the Content-* headers
    for h, v in msg.items():
        if h.lower().startswith('content-'):
            inner[h] = v
    inner.set_payload(msg.get_payload())
    # For completeness
    inner.set_unixfrom(msg.get_unixfrom())
    inner.preamble = msg.preamble
    inner.epilogue = msg.epilogue
    # Don't copy get_charset, as this might be None, even if
    # get_content_charset isn't.  However, do make sure there is a default
    # content-type, even if the original message was not MIME.
    inner.set_default_type(msg.get_default_type())
    # BAW: HACK ALERT.
    if hasattr(msg, '__version__'):
        inner.__version__ = msg.__version__
    # Now, play games with the outer message to make it contain three
    # subparts: the header (if any), the wrapped message, and the footer (if
    # any).
    payload = [inner]
    if header:
        mimehdr = MIMEText(header.encode(lcset), 'plain', lcset)
        mimehdr['Content-Disposition'] = 'inline'
        payload.insert(0, mimehdr)
    if footer:
        mimeftr = MIMEText(footer.encode(lcset), 'plain', lcset)
        mimeftr['Content-Disposition'] = 'inline'
        payload.append(mimeftr)
    msg.set_payload(payload)
    del msg['content-type']
    del msg['content-transfer-encoding']
    del msg['content-disposition']
    msg['Content-Type'] = 'multipart/mixed'



def decorate(mlist, uri, extradict=None):
    """Expand the decoration template."""
    if uri is None:
        return ''
    # Get the decorator template.
    loader = getUtility(ITemplateLoader)
    template_uri = expand(uri, dict(
        listname=mlist.fqdn_listname,
        language=mlist.preferred_language.code,
        ))
    template = loader.get(template_uri)
    # Create a dictionary which includes the default set of interpolation
    # variables allowed in headers and footers.  These will be augmented by
    # any key/value pairs in the extradict.
    substitutions = dict(
        fqdn_listname = mlist.fqdn_listname,
        list_name     = mlist.list_name,
        host_name     = mlist.mail_host,
        display_name  = mlist.display_name,
        listinfo_uri  = mlist.script_url('listinfo'),
        list_requests = mlist.request_address,
        description   = mlist.description,
        info          = mlist.info,
        )
    if extradict is not None:
        substitutions.update(extradict)
    text = expand(template, substitutions)
    # Turn any \r\n line endings into just \n
    return re.sub(r' *\r?\n', r'\n', text)



class Decorate:
    """Decorate a message with headers and footers."""

    implements(IHandler)

    name = 'decorate'
    description = _('Decorate a message with headers and footers.')

    def process(self, mlist, msg, msgdata):
        "See `IHandler`."""
        process(mlist, msg, msgdata)



# Copyright (C) 2008-2012 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <http://www.gnu.org/licenses/>.

"""Mailman test layers."""

# XXX 2012-03-23 BAW: Layers really really suck.  For example, the
# test_owners_get_email() test requires that both the SMTPLayer and LMTPLayer
# be set up, but there's apparently no way to do that and make zope.testing
# happy.  This causes no tests failures, but it does cause errors at the end
# of the full test run.  For now, I'll ignore that, but I do want to
# eventually get rid of the zope.test* dependencies and use something like
# testresources or some such.

from __future__ import absolute_import, unicode_literals

__metaclass__ = type
__all__ = [
    'ConfigLayer',
    'LMTPLayer',
    'MockAndMonkeyLayer',
    'RESTLayer',
    'SMTPLayer',
    'is_testing',
    ]


import os
import sys
import shutil
import logging
import datetime
import tempfile

from base64 import b64encode
from lazr.config import as_boolean, as_timedelta
from pkg_resources import resource_string
from textwrap import dedent
from urllib2 import Request, URLError, urlopen
from zope.component import getUtility

from mailman.config import config
from mailman.core import initialize
from mailman.core.initialize import INHIBIT_CONFIG_FILE
from mailman.core.logging import get_handler
from mailman.interfaces.domain import IDomainManager
from mailman.testing.helpers import (
    TestableMaster, get_lmtp_client, reset_the_world)
from mailman.testing.mta import ConnectionCountingController
from mailman.utilities.string import expand


TEST_TIMEOUT = datetime.timedelta(seconds=5)
NL = '\n'



class MockAndMonkeyLayer:
    """Layer for mocking and monkey patching for testing."""

    # Set this to True to enable predictable datetimes, uids, etc.
    testing_mode = False

    # A registration of all testing factories, for resetting between tests.
    _resets = []

    @classmethod
    def testTearDown(cls):
        for reset in cls._resets:
            reset()

    @classmethod
    def register_reset(cls, reset):
        cls._resets.append(reset)



class ConfigLayer(MockAndMonkeyLayer):
    """Layer for pushing and popping test configurations."""

    var_dir = None
    styles = None

    @classmethod
    def setUp(cls):
        # Set up the basic configuration stuff.  Turn off path creation until
        # we've pushed the testing config.
        config.create_paths = False
        initialize.initialize_1(INHIBIT_CONFIG_FILE)
        assert cls.var_dir is None, 'Layer already set up'
        # Calculate a temporary VAR_DIR directory so that run-time artifacts
        # of the tests won't tread on the installation's data.  This also
        # makes it easier to clean up after the tests are done, and insures
        # isolation of test suite runs.
        cls.var_dir = tempfile.mkdtemp()
        # We need a test configuration both for the foreground process and any
        # child processes that get spawned.  lazr.config would allow us to do
        # it all in a string that gets pushed, and we'll do that for the
        # foreground, but because we may be spawning processes (such as
        # runners) we'll need a file that we can specify to the with the -C
        # option.  Craft the full test configuration string here, push it, and
        # also write it out to a temp file for -C.
        test_config = dedent("""
        [mailman]
        layout: testing
        [passwords]
        password_scheme: cleartext
        [paths.testing]
        var_dir: %s
        [devmode]
        testing: yes
        """ % cls.var_dir)
        # Read the testing config and push it.
        test_config += resource_string('mailman.testing', 'testing.cfg')
        config.create_paths = True
        config.push('test config', test_config)
        # Initialize everything else.
        initialize.initialize_2()
        initialize.initialize_3()
        # When stderr debugging is enabled, subprocess root loggers should
        # also be more verbose.
        if cls.stderr:
            test_config += dedent("""
            [logging.root]
            propagate: yes
            level: debug
            """)
        # Enable log message propagation and reset the log paths so that the
        # doctests can check the output.
        for logger_config in config.logger_configs:
            sub_name = logger_config.name.split('.')[-1]
            if sub_name == 'root':
                continue
            logger_name = 'mailman.' + sub_name
            log = logging.getLogger(logger_name)
            log.propagate = True
            # Reopen the file to a new path that tests can get at.  Instead of
            # using the configuration file path though, use a path that's
            # specific to the logger so that tests can find expected output
            # more easily.
            path = os.path.join(config.LOG_DIR, sub_name)
            get_handler(sub_name).reopen(path)
            log.setLevel(logging.DEBUG)
            # If stderr debugging is enabled, make sure subprocesses are also
            # more verbose.
            if cls.stderr:
                test_config += expand(dedent("""
                [logging.$name]
                propagate: yes
                level: debug
                """), dict(name=sub_name, path=path))
        # zope.testing sets up logging before we get to our own initialization
        # function.  This messes with the root logger, so explicitly set it to
        # go to stderr.
        if cls.stderr:
            console = logging.StreamHandler(sys.stderr)
            formatter = logging.Formatter(config.logging.root.format,
                                          config.logging.root.datefmt)
            console.setFormatter(formatter)
            logging.getLogger().addHandler(console)
        # Write the configuration file for subprocesses and set up the config
        # object to pass that properly on the -C option.
        config_file = os.path.join(cls.var_dir, 'test.cfg')
        with open(config_file, 'w') as fp:
            fp.write(test_config)
            print >> fp
        config.filename = config_file

    @classmethod
    def tearDown(cls):
        assert cls.var_dir is not None, 'Layer not set up'
        config.pop('test config')
        shutil.rmtree(cls.var_dir)
        cls.var_dir = None

    @classmethod
    def testSetUp(cls):
        # Add an example domain.
        getUtility(IDomainManager).add(
            'example.com', 'An example domain.',
            'http://lists.example.com', 'postmaster@example.com')
        config.db.commit()

    @classmethod
    def testTearDown(cls):
        reset_the_world()

    # Flag to indicate that loggers should propagate to the console.
    stderr = False

    @classmethod
    def enable_stderr(cls):
        """Enable stderr logging if -e/--stderr is given.

        We used to hack our way into the zc.testing framework, but that was
        undocumented and way too fragile.  Well, this probably is too, but now
        we just scan sys.argv for -e/--stderr and enable logging if found.
        Then we remove the option from sys.argv.  This works because this
        method is called before zope.testrunner sees the options.

        As a bonus, we'll check an environment variable too.
        """
        if '-e' in sys.argv:
            cls.stderr = True
            sys.argv.remove('-e')
        if '--stderr' in sys.argv:
            cls.stderr = True
            sys.argv.remove('--stderr')
        if len(os.environ.get('MM_VERBOSE_TESTLOG', '').strip()) > 0:
            cls.stderr = True

    # The top of our source tree, for tests that care (e.g. hooks.txt).
    root_directory = None

    @classmethod
    def set_root_directory(cls, directory):
        """Set the directory at the root of our source tree.

        zc.recipe.testrunner runs from parts/test/working-directory, but
        that's actually changed over the life of the package.  Some tests
        care, e.g. because they need to find our built-out bin directory.
        Fortunately, buildout can give us this information.  See the
        `buildout.cfg` file for where this method is called.
        """
        cls.root_directory = directory



class SMTPLayer(ConfigLayer):
    """Layer for starting, stopping, and accessing a test SMTP server."""

    smtpd = None

    @classmethod
    def setUp(cls):
        assert cls.smtpd is None, 'Layer already set up'
        host = config.mta.smtp_host
        port = int(config.mta.smtp_port)
        cls.smtpd = ConnectionCountingController(host, port)
        cls.smtpd.start()

    @classmethod
    def tearDown(cls):
        assert cls.smtpd is not None, 'Layer not set up'
        cls.smtpd.clear()
        cls.smtpd.stop()

    @classmethod
    def testSetUp(cls):
        # Make sure we don't call our superclass's testSetUp(), otherwise the
        # example.com domain will get added twice.
        pass

    @classmethod
    def testTearDown(cls):
        cls.smtpd.reset()
        cls.smtpd.clear()



class LMTPLayer(ConfigLayer):
    """Layer for starting, stopping, and accessing a test LMTP server."""

    lmtpd = None

    @staticmethod
    def _wait_for_lmtp_server():
        get_lmtp_client(quiet=True)

    @classmethod
    def setUp(cls):
        assert cls.lmtpd is None, 'Layer already set up'
        cls.lmtpd = TestableMaster(cls._wait_for_lmtp_server)
        cls.lmtpd.start('lmtp')

    @classmethod
    def tearDown(cls):
        assert cls.lmtpd is not None, 'Layer not set up'
        cls.lmtpd.stop()
        cls.lmtpd = None

    @classmethod
    def testSetUp(cls):
        # Make sure we don't call our superclass's testSetUp(), otherwise the
        # example.com domain will get added twice.
        pass



class RESTLayer(SMTPLayer):
    """Layer for starting, stopping, and accessing the test REST layer."""

    server = None

    @staticmethod
    def _wait_for_rest_server():
        until = datetime.datetime.now() + as_timedelta(config.devmode.wait)
        while datetime.datetime.now() < until:
            try:
                request = Request('http://localhost:9001/3.0/system')
                basic_auth = '{0}:{1}'.format(config.webservice.admin_user,
                                              config.webservice.admin_pass)
                request.add_header('Authorization',
                                   'Basic ' + b64encode(basic_auth))
                fp = urlopen(request)
            except URLError:
                pass
            else:
                fp.close()
                break
        else:
            raise RuntimeError('REST server did not start up')

    @classmethod
    def setUp(cls):
        assert cls.server is None, 'Layer already set up'
        cls.server = TestableMaster(cls._wait_for_rest_server)
        cls.server.start('rest')

    @classmethod
    def tearDown(cls):
        assert cls.server is not None, 'Layer not set up'
        cls.server.stop()
        cls.server = None



def is_testing():
    """Return a 'testing' flag for use with the predictable factories.

    :return: True when in testing mode.
    :rtype: bool
    """
    return (MockAndMonkeyLayer.testing_mode or
            as_boolean(config.devmode.testing))
